#============================================================================#
#                             Importing Libraries                            #
#============================================================================#
import numpy as np                          # package for scientific computing
import matplotlib.pyplot as plt             # plotting library
#============================================================================#
#                                 RK4 Method                                 #
#============================================================================#
def rk4(f, x0, dt, t0):
    """
    Takes one RK4 step to solve a system of ODE equations of the form
    ``dx/dt = f(u,t)``,  ``x(t0) = x0``.

    Parameters
    ----------
    f : callable
        User-suplied function that evaluates the derivatives. Must be in
        the form ``f(x,t)``, where ``x`` is the argument in the form of a
        1-D array and ``t`` is the time.
    x0 : ndarray
        Solution estimated at current time.
    dt : float
        Time step size.
    t0 : float
        Current time.

    Returns
    -------
    out : ndarray
        The solution estimated at ``t0 + dt``.

    """
    x0 = np.asarray(x0)
    # K1
    k1 = f(x0, t0)
    # K2
    t = t0 + dt/2.0
    x = x0 + k1*dt/2.0
    k2 = f(x, t)
    # K3
    t = t0 + dt/2.0
    x = x0 + k2*dt/2.0
    k3 = f(x, t)
    # K4
    t = t0 + dt
    x = x0 + k3*dt
    k4 = f(x, t)

    return x0 + dt*(k1 + 2.0*(k2 + k3) + k4)/6.0
#============================================================================#
#                           Integration Function                             #
#============================================================================#
def integrate(t0, dt, n, x0, f):
    """
    Integrates an ODE system with time step `dt` from `t0` until `tf=t0+n*dt`.

    The 4th order Runge-Kutta method is used to integrate a system of
    equations of type:

        dx/dt  = f(x,t) ,  x0 = x(t0)

    Parameters
    ----------
    t0 : float
         Time at solution is known (initial time).
    dt : float
         Time step size.
    n : int
        Number of time step to evaluate.
    x0 : ndarray
         Estimated solution at t0.
    f : function object
           User-supplied function which returns the derivatives at `t`.

    Returns
    -------
    data : ndarray
           Containing array([ [t0, x0[:]], [t1, x1[:]], ..., [tn, xn[:]] ])

    """
    data = []
    x = x0
    t = t0
    data.append([t]+list(x))
    for i in range(n):
        x = rk4(f, x, dt, t)
        t = t + dt
        data.append([t]+list(x))

    return np.array(data)
#===========================================================================#
