# -*- coding: utf-8 -*-
#============================================================================#
#                          The simple pendulum problem                       #
#============================================================================#
from numpy import sin, cos
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.animation import PillowWriter, HTMLWriter
#============================================================================#
#                                 RK4 Method                                 #
#============================================================================#
def rk4(func, x0, dt, t0):
    """
    Takes one RK4 step to solve a system of ODE equations of the form
    ``dx/dt = f(x,t)``,  ``x(t0) = x0``.

    Parameters
    ----------
    func :: callable
        User-suplied function that evaluates the derivatives. Must be in
        the form ``f(x,t)``, where ``x`` is the argument in the form of a
        1-D array and ``t`` is the time.
    x0 :: ndarry
        Solution estimated at current time.
    dt :: float
        Time step size.
    t0 :: float
        Current time.

    Returns
    -------
    out :: ndarray
        The solution estimated at ``t0 + dt``.

    """
    x0 = np.asarray(x0)
    # K1
    k1 = func(x0, t0)
    # K2
    t = t0 + dt/2.0
    x = x0 + k1*dt/2.0
    k2 = func(x, t)
    # K3
    t = t0 + dt/2.0
    x = x0 + k2*dt/2.0
    k3 = func(x, t)
    # K4
    t = t0 + dt
    x = x0 + k3*dt
    k4 = func(x, t)

    return x0 + dt*(k1 + 2.0*(k2 + k3) + k4)/6.0
#============================================================================#
#                           Integration Function                             #
#============================================================================#
def integrate(t0, dt, n, x0, func):
    """
    Integrates an ODE system with time step `dt` from `t0` until `tf=t0+n*dt`.

    The 4th order Runge-Kutta method is used to integrate a system of
    equations of type:

        dx/dt  = f(x,t) ,  x0 = x(t0)

    Parameters
    ----------
    t0 : float
         Time at solution is known (initial time).
    dt : float
         Time step size.
    n : int
        Number of time step to evaluate.
    x0 : ndarray
         Estimated solution at t0.
    func : function object
           User-supplied function which returns the derivatives at `t`.

    Returns
    -------
    data : ndarray
           Containing array([ [t0, x0[:]], [t1, x1[:]], ..., [tn, xn[:]] ])

    """
    data = []
    x = np.copy(x0)
    t = t0
    data.append([t]+list(x))
    for i in range(n):
        x = rk4(func, x, dt, t)
        t = t + dt
        data.append([t]+list(x))

    return np.array(data) 
#============================================================================#
# Right-hand side of linear simple pendulum differential equations           #
#============================================================================#
def lin_sim_pen(state, t):

    ff = np.zeros_like(state)
    ff[0] = state[1]
    ff[1] = amp*sin(omega*t) - ksi*state[1] -(G/L)*state[0]
    return ff
#============================================================================#
# Right-hand side of nonlinear simple pendulum differential equations        #
#============================================================================#
def nonlin_sim_pen(state, t):

    ff = np.zeros_like(state)
    ff[0] = state[1]
    ff[1] = amp*sin(omega*t) - ksi*state[1] -(G/L)*np.sin(state[0])
    return ff
#============================================================================#
#
#============================================================================#
def plots(y, X1, X2, label, ls):
        
    t     = y[:,0]                      # time array
    theta = y[:,1]                      # angle
    w     = y[:,2]                      # angular velocity
    theta = np.degrees(theta)           # converts from rad to degree
    w = np.degrees(w)                   # converts from rad to degree
    #========================================================================#
    # Save to file                                                           #
    #========================================================================#
#    fobj = open('output_simple_pendulum.dat', 'w')
#    fobj.write('   Tempo       X1         X2 \n')
#    for i in range(len(t)):
#        fobj.write(' %f   %f   %f \n'%(t[i],X1[i],X2[i]))
#    fobj.close()
    #========================================================================#
    # Plot results - Figure 1                                                #
    #========================================================================#
    plt.figure(1)
    plt.plot(t, theta, ls, label=label, alpha=1.0)
    plt.xlabel(' Tempo (s) ')
    plt.ylabel('$\\theta [\degree]$')
    plt.title(' Pêndulo Simples')
    plt.xlim(0.0, 50.0)
    plt.ylim(-1.5*np.max(theta), 1.5*np.max(theta))
    plt.legend()
    #========================================================================#
    # State space                                                            #
    #========================================================================#
    plt.figure(2)
    plt.plot(theta, w, ls, label=label)
    plt.xlabel('$\\theta [\degree]$')
    plt.ylabel('$\\omega [\degree/s]$')
    plt.title(' Pêndulo Simples')
    plt.legend()
#============================================================================#
#                               MAIN PROGRAM                                 #
#============================================================================#
if __name__=='__main__':
    #========================================================================#
    # Inputs                                                                 #
    #========================================================================#
    G  = 9.8      # acceleration due to gravity, in m/s^2
    L  = 1.0      # length of pendulum, in m
    ksi = 0.2     # damping ratio
    amp = 0.0     # excitation ampltitude
    omega = 12.   # excitation frequency
    t0 = 0.0      # initial time
    dt = 0.05     # time step
    N  = 1000     # number of time steps
    
    th1 = -120.0   # initial angle (degrees)
    w1 =   500.0   # initial angular velocity (degrees per second)
    init_conditions = np.radians([th1, w1])  # converts to radians
    plt.close('all')
    #========================================================================#
    # Linear Simple Pendulum Solution using RK4                              #
    #========================================================================#
    y = integrate(0., dt, N, init_conditions, lin_sim_pen)  # matrix with results
    # converts to cartesian coordenates    
    X1_lin = -L*sin(y[:, 1])
    X2_lin = L*cos(y[:, 1])
    plots(y, X1_lin, X2_lin, 'linear', '-b')
    #========================================================================#
    # Nonlinear Simple Pendulum Solution using RK4                           #
    #========================================================================#
    y = integrate(0., dt, N, init_conditions, nonlin_sim_pen)  # matrix with results
    X1_non = -L*sin(y[:, 1])
    X2_non = L*cos(y[:, 1])
    plots(y, X1_non, X2_non, 'não-linear','-r')    
    plt.show()
#============================================================================#